//-------------------------------------------------------------------------
// Fichero:  anaritex.c
// Autor:    Pedro Reina <pedro@pedroreina.net>
// Fecha:    X.24.4.2024
// Objetivo: Averiguar si es posible obtener un número Meta usando N números
//           (cada uno una vez o ninguna) y las cuatro operaciones elementales;
//           si no se puede, el programa debe dar el número más aproximado
//           posible (por exceso o por defecto). En cualquier caso se debe dar
//           la operación que hay que realizar para obtener el número con la
//           menor cantidad de operaciones
// Entradas: Los N números y el número Meta
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Licencia
//
// (c) 2024 Pedro Reina y EQ Sistemas Inteligentes
//
// This program is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation, version 2
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Ficheros de cabecera
//-------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//-------------------------------------------------------------------------
// Definición de constantes
//-------------------------------------------------------------------------

// El mayor número de datos de partida que admitimos
#define MAX_DATO 10

// Los dos valores lógicos
#define SI 1
#define NO 0

// Para indicar el fin de una cadena
#define NULO '\0'

// Las cuatro operaciones
#define SUMA     0
#define RESTA    1
#define PRODUCTO 2
#define COCIENTE 3

// El separador para la notación polaca inversa
#define SEPARADOR ","

// Un número entero bastante grande
#define MAXVALOR 1073741824

//-------------------------------------------------------------------------
// Definición de tipos generales
//-------------------------------------------------------------------------

// Para valores lógicos SI y NO
typedef char logico;

// Para valores enteros que pueden ser grandes
typedef unsigned long entero;

// Para la expresión de las operaciones
typedef char cadena[6*MAX_DATO];

// Para números enteros que sean pequeños
typedef short int contador;

//-------------------------------------------------------------------------
// Definición del objeto Candidato (Can)
//     Un candidato es una posible solución del problema
//-------------------------------------------------------------------------

//--------------------
// Estructura de datos
//--------------------

typedef struct {
    // Un número, que podría ser la solución
    entero Numero;

    // Las operaciones para llegar al número
    cadena Camino;

    // Cuántas operaciones hacen falta para llegar al número
    contador Esfuerzo;
    } can_st, *candidato;

//---------------------
// Definición de macros
//---------------------

#define Can_Numero(c) ((c)->Numero)
#define Can_Camino(c) ((c)->Camino)
#define Can_Esfuerzo(c) ((c)->Esfuerzo)
#define Can_PonNumero(c,n) (((c)->Numero)=(n))
#define Can_PonCamino(c,m) (((c)->Camino)=(m))
#define Can_PonEsfuerzo(c,e) (((c)->Esfuerzo)=(e))

//-------------------------
// Declaración de funciones
//-------------------------

// Crea un candidato
candidato Can_Crea(void);

// Destruye un candidato
void Can_Destruye(candidato Candidato);

// Copia los datos de un candidato a otro
void Can_Copia(candidato Destino, candidato Origen);

//-------------------------------------------------------------------------
// Definición del objeto Nodo (Nod)
//     Un nodo es un conjunto de candidatos que provienen del mismo padre
//-------------------------------------------------------------------------

//--------------------
// Estructura de datos
//--------------------

typedef struct {
    // La cantidad de candidatos del nodo
    contador Longitud;

    // Un vector con los candidatos
    candidato *Vector;
    } nod_st, *nodo;

//---------------------
// Definición de macros
//---------------------

#define Nod_Longitud(n) ((n)->Longitud)
#define Nod_Vector(n) ((n)->Vector)
#define Nod_PonLongitud(n,g) (((n)->Longitud)=(g))
#define Nod_PonVector(n,t) (((n)->Vector)=(t))

//-------------------------
// Declaración de funciones
//-------------------------

// Crea un nodo
nodo Nod_Crea(contador Long);

// Destruye un nodo
void Nod_Destruye(nodo Nodo);

// Da un candidato de un nodo
candidato Nod_Candidato(nodo Nodo, contador i);

// Pone un candidato en un nodo
void Nod_PonCandidato(nodo Nodo, contador i, candidato Candidato);

// Crea un nodo hijo
nodo Nod_Hijo(nodo Padre, contador i1, contador i2, candidato Candidato);

//-------------------------------------------------------------------------
// Definición del objeto Calculadora (Cal)
//     La calculadora convierte de RPN a notación algebraica
//-------------------------------------------------------------------------

//--------------------
// Estructura de datos
//--------------------

// Para almacenar los números
entero Cal_Pila[MAX_DATO+1];

//---------------------
// Definición de macros
//---------------------

//-------------------------------------------------------------------------
// Macro:    Cal_EsOperacion()
// Objetivo: Decidir si un carácter representa una operación
// Entradas: El carácter
// Salidas:  Lógica: SI si lo es, NO en caso contrario
//-------------------------------------------------------------------------

#define Cal_EsOperacion(c) ((c=='+')||(c=='-')||(c=='*')||(c=='/'))

//-------------------------------------------------------------------------
// Macro:    Cal_EsSeparador()
// Objetivo: Decidir si un carácter es el separador
// Entradas: El carácter
// Salidas:  Lógica: SI si lo es, NO en caso contrario
//-------------------------------------------------------------------------

#define Cal_EsSeparador(c) (c==SEPARADOR[0])

//-------------------------
// Declaración de funciones
//-------------------------

// Prepara la calculadora para trabajar
void Cal_Inicia(void);

// Realiza una operación
entero Cal_Opera(char);

// Inserta un número en la calculadora
entero Cal_Inserta(entero);

// Explica la solución en notación algebraica
void Cal_Explica(char *);

//-------------------------------------------------------------------------
// Definición de macros de manejo de cadenas (Cad)
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Macro:    Cad_ConvierteEntero()
// Objetivo: Convertir un número entero en una cadena
// Entradas: La cadena y el número
// Salidas:  La cadena
//-------------------------------------------------------------------------

#define Cad_ConvierteEntero(c,n) sprintf((c),"%ul",(n))

//-------------------------------------------------------------------------
// Macro:    Cad_Entero()
// Objetivo: Convertir una cadena en un número entero
// Entradas: La cadena
// Salidas:  El número
//-------------------------------------------------------------------------

#define Cad_Entero(c) ((entero)(atoi(c)))

//-------------------------------------------------------------------------
// Macro:    Cad_Copia()
// Objetivo: Copiar una cadena en otra
// Entradas: La cadena destino y la cadena origen
// Salidas:  La cadena destino
//-------------------------------------------------------------------------

#define Cad_Copia(d,o) strcpy((d),(o))

//-------------------------------------------------------------------------
// MACRO:    Cad_Une()
// OBJETIVO: Unir una cadena con otra
// ENTRADAS: La cadena destino y la cadena origen
// SALIDAS:  La cadena destino
//-------------------------------------------------------------------------

#define Cad_Une(d,o) strcat((d),(o))

//-------------------------------------------------------------------------
// Declaración de funciones de manejo de cadenas (Cad)
//-------------------------------------------------------------------------

// Extrae una cadena de otra
char *Cad_Trozo(cadena Destino, cadena Origen,
                 contador Comienzo, contador Fin);

//-------------------------------------------------------------------------
// Definición de macros generales
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Macro:     Distancia()
// Objetivo:  Calcular la distancia entre dos enteros
// Entradas:  Los dos enteros
// Salidas:   La diferencia
// Algoritmo: Devolver la diferencia entre el mayor y el menor
//-------------------------------------------------------------------------

#define Distancia(a,b) ( (a)>(b) ? ((a)-(b)) : ((b)-(a)) )

//-------------------------------------------------------------------------
// Declaración de funciones generales
//-------------------------------------------------------------------------

// Explica cuáles son los datos de partida
void ExplicaDatos(void);

// Explica cuál es el número Meta
void ExplicaMeta(void);

// Resuelve el problema
void Resuelve(void);

// Busca soluciones a partir de un nodo
void Busca (nodo Nodo);

// Compueba si un candidato es solución
void Comprueba (candidato Candidato);

// Explica la mejor aproximación obtenida
void ExplicaNumero(void);

// Une dos caminos con una operación
char *ConcatenaCaminos();

//-------------------------------------------------------------------------
// Variables globales
//-------------------------------------------------------------------------

// Con cuántos datos trabajaremos
contador TotalDatos;

// Los datos de partida
entero Datos[MAX_DATO];

// El número que hay que alcanzar
entero Meta;

// La mejor solución hasta el momento
candidato Solucion;

// La diferencia entre Meta y Solucion
entero Diferencia = MAXVALOR;

// El menor número de operaciones para calcular la solución
entero MenorEsfuerzo = MAXVALOR;

// El número total de operaciones realizadas
entero TotalOperaciones = 0;

//-------------------------------------------------------------------------
// Programa principal
//
// Entradas: Por la línea de órdenes recibimos los N números y el número Meta
// Salidas:  Confirmación de los datos introducidos
//           El número más aproximado que se haya encontrado
//           Las operaciones para obtenerlo
//           El número de operaciones realizadas
//-------------------------------------------------------------------------

int main (int Narg, char *Arg[])
    {
    contador i;

    // Nos presentamos
    printf ("Analizador aritmético exhaustivo\n");

    // Si no hay datos suficientes, nos quejamos
    if ( Narg < 5 )
        { printf ("Uso: ./anaritex dato1 ... datoN Meta\n"); }

    // Si hay demasiados, nos quejamos también
    else if ( Narg > MAX_DATO+2 )
        { printf ("Error: el número máximo de datos es %d\n", MAX_DATO); }

    // Leemos la línea de órdenes
    else
        {
        // Tomamos los datos
        for ( i = 1; i < Narg-1; i++ )
            { Datos[i-1] = Cad_Entero(Arg[i]); }
        TotalDatos = i-1;
        ExplicaDatos();

        // Tomamos el número meta
        Meta = Cad_Entero(Arg[TotalDatos+1]);
        ExplicaMeta();
        }

    // Resolvemos
    Solucion = Can_Crea();
    printf ("Analizando...\n");
    Resuelve();

    // Informamos
    ExplicaNumero();
    Cal_Explica(Can_Camino(Solucion));
    printf ("Operaciones analizadas: %u\n", TotalOperaciones);

    // Todo ha ido bien (supuestamente)
    return (0);
    }

//-------------------------------------------------------------------------
// Funciones generales
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Función:  ExplicaDatos()
// Objetivo: Dar por pantalla los datos de partida
// Entradas: Variables globales Datos y TotalDatos
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void ExplicaDatos (void)
    {
    contador i;
    printf ("Números de partida: ");
    for ( i = 0; i < TotalDatos; i++ )
        { printf ("%d ", Datos[i]); }
    printf ("\n");
    }

//-------------------------------------------------------------------------
// Función:  ExplicaMeta()
// Objetivo: Dar por pantalla el número meta
// Entradas: Variable global Meta
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void ExplicaMeta (void)
    { printf ("Número objetivo: %d\n", Meta); }

//-------------------------------------------------------------------------
// Función:  ExplicaNumero()
// Objetivo: Dar por pantalla el mejor número encontrado
// Entradas: Variable global Solucion
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void ExplicaNumero (void)
    {
    char *Exacto;
    entero Num;

    Num = Can_Numero(Solucion);
    Exacto = (Num == Meta) ? "¡Exacto!" : "";
    printf ("Mejor aproximación encontrada: %d %s\n", Num, Exacto);
    }

//-------------------------------------------------------------------------
// Función:  Resuelve()
// Objetivo: Resolver el problema
// Entradas: Variables globales Datos y TotalDatos
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void Resuelve (void)
    {
    contador i;
    nodo NodoRaiz;
    candidato Candidato;

    // Preparamos el nodo raíz
    NodoRaiz = Nod_Crea(TotalDatos);
    for ( i=0 ; i<TotalDatos ; i++ )
        {
        Candidato = Can_Crea();
        Can_PonNumero(Candidato, Datos[i]);
        Cad_ConvierteEntero(Can_Camino(Candidato),Datos[i]);
        Can_PonEsfuerzo(Candidato, 0);
        Nod_PonCandidato(NodoRaiz, i, Candidato);
        }

    // Lanzamos recursivamente la búsqueda de soluciones
    Busca(NodoRaiz);
    }

//-------------------------------------------------------------------------
// Función:  Busca (nodo Nodo)
// Objetivo: Buscar soluciones a partir de un nodo
// Entradas: Un nodo
// Salidas:  Ninguna. Las variables globales pueden quedar modificadas
//-------------------------------------------------------------------------
void Busca (nodo Nodo)
    {
    contador i, j, Operacion, Esf1, Esf2;
    candidato Candidato1, Candidato2, NuevoCandidato;
    nodo NodoHijo;
    entero Num1, Num2, NuevoNum;
    logico HayQueCalcular;
    char *Cam1, *Cam2, *NuevoCam;
    cadena Destino;

    // Comprobamos cada candidato por si nos vale como solución
    contador Long = Nod_Longitud(Nodo);
    for ( i=0 ; i<Long ; i++ )
        { Comprueba(Nod_Candidato(Nodo, i)); }

     // Probamos todas las parejas de candidatos
    for ( i=0   ; i<Long-1 ; i++ ) {
    for ( j=i+1 ; j<Long   ; j++ )
        {
        // Los dos candidatos y sus datos
        Candidato1 = Nod_Candidato(Nodo, i);
        Candidato2 = Nod_Candidato(Nodo, j);
        Num1 = Can_Numero(Candidato1);
        Num2 = Can_Numero(Candidato2);
        Cam1 = Can_Camino(Candidato1);
        Cam2 = Can_Camino(Candidato2);
        Esf1 = Can_Esfuerzo(Candidato1);
        Esf2 = Can_Esfuerzo(Candidato2);

        // Probamos todas las operaciones posibles
        for ( Operacion=SUMA ; Operacion<=COCIENTE ; Operacion++ )
            {
            HayQueCalcular = NO;
            switch ( Operacion )
                {
                case SUMA:
                    NuevoNum = Num1+Num2;
                    ConcatenaCaminos (Destino, Cam1, "+", Cam2);
                    HayQueCalcular = SI;
                    break;

                case RESTA:
                    if ( Num1 > Num2 )
                        {
                        NuevoNum = Num1-Num2;
                        ConcatenaCaminos (Destino, Cam1, "-", Cam2);
                        HayQueCalcular = SI;
                        }
                    if ( Num1 < Num2 )
                        {
                        NuevoNum = Num2-Num1;
                        ConcatenaCaminos (Destino, Cam2, "-", Cam1);
                        HayQueCalcular = SI;
                        }
                    break;

                case PRODUCTO:
                    if ( Num1 != 1 && Num2 != 1)
                        {
                        NuevoNum = Num1*Num2;
                        ConcatenaCaminos (Destino, Cam1, "*", Cam2);
                        HayQueCalcular = SI;
                        }
                    break;

                case COCIENTE:
                    if ( Num1 % Num2 == 0 && Num2 != 1)
                        {
                        NuevoNum = Num1/Num2;
                        ConcatenaCaminos (Destino, Cam1, "/", Cam2);
                        HayQueCalcular = SI;
                        }
                    if ( Num2 % Num1 == 0 && Num1 != 1)
                        {
                        NuevoNum = Num2/Num1;
                        ConcatenaCaminos (Destino, Cam2, "/", Cam1);
                        HayQueCalcular = SI;
                        }
                    break;
                } // Fin del switch

            if ( HayQueCalcular )
                {
                // Una operación más
                TotalOperaciones++;

                // Preparamos el nuevo candidato
                NuevoCandidato = Can_Crea();
                Can_PonNumero(NuevoCandidato, NuevoNum);
                Cad_Copia(Can_Camino(NuevoCandidato),Destino);
                Can_PonEsfuerzo(NuevoCandidato, Esf1+Esf2+1);

                // Creamos un nodo hijo
                NodoHijo = Nod_Hijo(Nodo, i, j, NuevoCandidato);

                // Buscamos a partir del nuevo nodo
                Busca(NodoHijo);
                }

            } // Fin del for Operacion

        }} // Fin de los for i,j

    // Ya hemos estudiado el nodo, liberamos su memoria
    Nod_Destruye(Nodo);
    }

//-------------------------------------------------------------------------
// Función:  Comprueba(candidato Candidato)
// Objetivo: Comprobar si un candidato es solución
// Entradas: Un candidato
// Salidas:  Ninguna. Las variables globales pueden quedar modificadas
//-------------------------------------------------------------------------

void Comprueba(candidato Candidato)
    {
    entero Numero, DiferenciaPropuesta;
    contador Esfuerzo;

    // Tomamos los datos del candidato
    Numero = Can_Numero(Candidato);
    Esfuerzo = Can_Esfuerzo(Candidato);

    // Comparamos
    DiferenciaPropuesta = Distancia(Numero,Meta);

    // Nos quedamos el candidato si es mejor que lo que llevamos
    if ( DiferenciaPropuesta < Diferencia ||
         (DiferenciaPropuesta == Diferencia && Esfuerzo < MenorEsfuerzo) )
        {
        Can_Copia(Solucion, Candidato);
        Diferencia = DiferenciaPropuesta;
        MenorEsfuerzo = Esfuerzo;
        }
    }

//-------------------------------------------------------------------------
// Función:  ConcatenaCaminos()
// Objetivo: Unir dos caminos con el signo de una operación en RPN
// Entradas: Cadena destino, primer camino, operación, y segundo camino
// Salidas:  La cadena resultante de unir todo
// Algoritmo: Copiar en Aux el separador
//            Añadir a Aux el segundo camino
//            Añadir a Aux el separador
//            Añadir a Aux la operacion
//            Copiar en Destino el primer camino
//            Añadir a Destino Aux
//            Devolver Destino
//-------------------------------------------------------------------------

char *ConcatenaCaminos(Destino,Primero,Operacion,Segundo)
char Destino[];
char Primero[];
char Operacion[];
char Segundo[];
  {
  cadena Aux;

  Cad_Copia (Aux,SEPARADOR);
  Cad_Une (Aux,Segundo);
  Cad_Une (Aux,SEPARADOR);
  Cad_Une (Aux,Operacion);
  Cad_Copia (Destino,Primero);
  Cad_Une (Destino,Aux);

  return (Destino);
  }

//-------------------------------------------------------------------------
// Funciones de manejo de cadenas (Cad)
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Función:  Cad_Trozo()
// Objetivo: Extraer una cadena de otra
// Entradas: Cadena destino, cadena origen y puntos de comienzo y fin del
//           corte
// Salidas:  La cadena destino queda modificada y se devuelve
// Algoritmo: Copiar los caracteres entre comienzo y fin y añadir NULO
//-------------------------------------------------------------------------

char *Cad_Trozo (cadena Destino, cadena Origen,
                 contador Comienzo, contador Fin)
  {
  contador i;

  for ( i=Comienzo ; i<=Fin ; i++ )
    { Destino[i-Comienzo] = Origen[i]; }

  Destino[i-Comienzo] = NULO;

  return (Destino);
  }

//-------------------------------------------------------------------------
// Funciones del objeto Candidato (Can)
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Función:  Can_Crea()
// Objetivo: Crear un candidato en memoria
// Entradas: Ninguna
// Salidas:  El candidato creado
//-------------------------------------------------------------------------

candidato Can_Crea(void)
    {
    candidato Nuevo;

    // Pedimos memoria al sistema
    Nuevo = (candidato) (malloc(sizeof(can_st)));

    // Avisamos si no se ha podido crear
    if ( !Nuevo )
        { printf ("Error: no hay memoria para crear un candidato\n"); }

    return ( Nuevo );
    }

//-------------------------------------------------------------------------
// Función:  Can_Destruye(candidato Candidato)
// Objetivo: Libera la memoria de un candidato
// Entradas: El candidato
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void Can_Destruye(candidato Candidato)
    {
    if ( Candidato )
        { free(Candidato); }
    }

//-------------------------------------------------------------------------
// Función:  Can_Copia(candidato Destino, candidato Origen)
// Objetivo: Copia los datos de un candidato en otro
// Entradas: El destino y el origen
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void Can_Copia(candidato Destino, candidato Origen)
    {
    Can_PonNumero(Destino,Can_Numero(Origen));
    Can_PonEsfuerzo(Destino,Can_Esfuerzo(Origen));
    Cad_Copia(Can_Camino(Destino),Can_Camino(Origen));
    }

//-------------------------------------------------------------------------
// Funciones del objeto Nodo (Nod)
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Función:  Nod_Crea(contador Long)
// Objetivo: Crear un nodo en memoria
// Entradas: La longitud del nodo
// Salidas:  El nodo creado
//-------------------------------------------------------------------------

nodo Nod_Crea(contador Long)
    {
    nodo Nuevo;
    candidato *Vector;

    // Pedimos memoria al sistema para el nodo
    Nuevo = (nodo) (malloc(sizeof(nod_st)));

    // Avisamos si no se ha podido crear
    if ( !Nuevo )
        { printf ("Error: no hay memoria para crear un nodo\n"); }

    // Anotamos la longitud del nodo
    Nod_PonLongitud (Nuevo, Long);

    // Pedimos memoria al sistema para el vector de candidatos
    Vector = malloc(Long*sizeof(candidato));

    // Avisamos si no se ha podido crear
    if ( !Vector )
        { printf ("Error: no hay memoria para crear un nodo\n"); }

    // Anotamos el vector del nodo
    Nod_PonVector (Nuevo, Vector);

    return ( Nuevo );
    }

//-------------------------------------------------------------------------
// Función:  Nod_Destruye()
// Objetivo: Libera la memoria de un nodo
// Entradas: Un nodo
// Salidas:  Ninguna
//-------------------------------------------------------------------------

void Nod_Destruye(nodo Nodo)
    {
    contador i, Long;

    // Pedimos la longitud del nodo
    Long = Nod_Longitud(Nodo);

    // Destruimos todos los candidatos
    for ( i=0 ; i<Long ; i++ )
        { Can_Destruye(Nod_Candidato(Nodo,i)); }

    // Liberamos la memoria del nodo
    free(Nodo);
    }

//-------------------------------------------------------------------------
// Función:  Nod_Candidato(nodo Nodo, contador i)
// Objetivo: Dar el candidato de lugar i
// Entradas: El nodo y el índice
// Salidas:  El candidato
//-------------------------------------------------------------------------

candidato Nod_Candidato(nodo Nodo, contador i)
    {
    candidato Candidato, *Vector;

    // Pedimos el vector del nodo
    Vector = Nod_Vector(Nodo);

    // Pedimos el candidato de lugar i
    Candidato = Vector[i];

    // Lo devolvemos
    return ( Candidato );
    }

//-------------------------------------------------------------------------
// Función:   Nod_PonCandidato(Nodo,i,Candidato)
// Objetivo:  Poner el candidato de lugar i
// Entradas:  El nodo y el índice
// Salidas:   Ninguna
//-------------------------------------------------------------------------

void Nod_PonCandidato(nodo Nodo, contador i, candidato Candidato)
    {
    candidato *Vector;

    // Pedimos el vector del nodo
    Vector = Nod_Vector(Nodo);

    // Anotamos el candidato
    Vector[i] = Candidato;
    }

//-------------------------------------------------------------------------
// Función:  Nod_Hijo(Padre,i1,i2,Candidato)
// Objetivo: Crear un nodo hijo quitando dos candidatos y añadiendo otro
// Entradas: El nodo padre, los índices de los candidatos que no se copian
//           y un candidato nuevo
// Salidas:  El nodo
//-------------------------------------------------------------------------

nodo Nod_Hijo(nodo Padre, contador i1, contador i2, candidato Candidato)
    {
    contador i, j, Long;
    candidato Nuevo;
    nodo Hijo;

    // Tomamos datos del nodo padre
    Long = Nod_Longitud(Padre);

    // Creamos un nuevo nodo
    Hijo = Nod_Crea(Long-1);

    // Copiamos candidatos de padre a hijo
    j = 0;
    for ( i=0 ; i<Long ; i++ )
        {
        if ( i != i1 && i != i2 )
            {
            Nuevo = Can_Crea();
            Nod_PonCandidato(Hijo, j, Nuevo);
            Can_Copia(Nod_Candidato(Hijo,j++), Nod_Candidato(Padre,i));
            }
        }

    // Añadimos el nuevo candidato
    Nod_PonCandidato(Hijo,j,Candidato);

    // Devolvemos el nodo hijo
    return ( Hijo );
    }

//-------------------------------------------------------------------------
// Funciones del objeto Calculadora (Cal)
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Función:   Cal_Inicia()
// Objetivo:  Dar valores iniciales a las pilas de la calculadora
// Entradas:  Ninguna
// Salidas:   La variable global Cal_Pila queda modificada
// Algoritmo: Poner todos los números a cero
//-------------------------------------------------------------------------

void Cal_Inicia (void)
  {
  contador i;

  for ( i=0 ; i<MAX_DATO ; i++ )
    { Cal_Pila[i] = 0; }
  }

//-------------------------------------------------------------------------
// Función:  Cal_Explica(char *Expresion)
// Entradas: Una cadena en notación RPN
// Salidas:  Ninguna
// Algoritmo:
//     Hacer Comienzo 0
//     Hacer Operacion NO
//     Iniciar la calculadora
//     Para cada carácter de la expresión, de índice i
//       Si es una operación
//         Hacer Cal_Opera() con ella
//         Hacer Operación SI
//         Hacer Comienzo i+1
//       Si es el separador
//         Si Operacion es NO
//           Hacer Fin i-1
//           Extraer el número de la expresión
//           Hacer Cal_Inserta() con el número
//         Hacer Comienzo i+1
//       Si es una cifra
//         Hacer Operacion NO
//     Devolver la cadena
//-------------------------------------------------------------------------

void Cal_Explica (char *Expresion)
  {
  contador Comienzo, Fin, i;
  logico   Operacion;
  char     c;
  entero   Num;
  cadena   Aux;

  Comienzo = 0;
  Operacion = NO;
  Cal_Inicia();


  printf ("Solución:\n");

  for ( i=0 ; c=Expresion[i] ; i++ )
    {
    if ( Cal_EsOperacion(c) )
             {
             Cal_Opera(c);
             Operacion = SI;
             Comienzo = i+1;
             }
    else if ( Cal_EsSeparador(c) )
             {
             if ( !Operacion )
               {
               Fin = i-1;
               Cad_Trozo (Aux,Expresion,Comienzo,Fin);
               Num = Cad_Entero (Aux);
               Cal_Inserta (Num);
               }
             Comienzo = i+1;
             }
    else
             { Operacion = NO; }
    }
  }

//-------------------------------------------------------------------------
// Función:  Cal_Inserta(entero Num)
// Objetivo: Insertar un número en las pilas
// Entradas: El número
// Salidas:  La variable global Cal_Pila queda modificada
//           Se devuelve el número insertado
// Algoritmo: Mover todos los niveles hacia arriba
//            Poner el número en el nivel más bajo
// Nota: Se pierde el valor más alto de la pila
//-------------------------------------------------------------------------

entero Cal_Inserta (entero Num)
  {
  contador i;

  for ( i=MAX_DATO-1 ; i>0 ; i-- )
    { Cal_Pila[i] = Cal_Pila[i-1]; }

  Cal_Pila[0] = Num;

  return ( Num );
  }

//-------------------------------------------------------------------------
// Función:  Cal_Opera()
// Objetivo: Realizar una operación
// Entradas: Un carácter indicando la operación (+,-,*,/)
// Salidas:  La variable global Cal_Pila queda modificada
//           Se devuelve el número calculado
// Algoritmo: Realizar la operación entre los niveles más bajos
//            Escribir el resultado en el nivel más bajo
//            Mostrar la operación
//            Bajar los niveles
// Nota: Se duplica el valor más alto de la pila
//-------------------------------------------------------------------------

entero Cal_Opera (char Operacion)
  {
  contador i;
  entero Num;

  switch ( Operacion )
    {
    case '+':
      Num = Cal_Pila[1]+Cal_Pila[0];
      break;
    case '-':
      Num = Cal_Pila[1]-Cal_Pila[0];
      break;
    case '*':
      Num = Cal_Pila[1]*Cal_Pila[0];
      break;
    case '/':
      Num = Cal_Pila[1]/Cal_Pila[0];
      break;
    }

  printf ( "    %lu %c %lu = %lu\n",
           Cal_Pila[1], Operacion, Cal_Pila[0], Num );

  Cal_Pila[0] = Num;

  for ( i=1 ; i<MAX_DATO-1 ; i++ )
    { Cal_Pila[i] = Cal_Pila[i+1]; }

  return ( Num );
  }
