#!/usr/bin/python
# coding: utf-8

#--------------------------------------------------------------------
# Fichero:  anaritex.py
# Objetivo: Analizador aritmético exhaustivo
# Autor:    Pedro Reina <pedro@pedroreina.net>
# Fecha:    V.26.4.2024
# Python:   comprobado con las versiones 2.7.3, 2.7.16 y 3.7.3
#--------------------------------------------------------------------

#--------------------------------------------------------------------
# Licencia
#
# (c) 2024 Pedro Reina y EQ Sistemas Inteligentes
#
# This program is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
#--------------------------------------------------------------------

#------------------------------------------------
# Estructuras de datos
#
# Un candidato es una terna con estos tres elementos:
#     Un número, que podría ser la solución
#     Una cadena que explica en RPN cómo calcular el número
#     Un número que indica cuántas operaciones se han usado
#
# Un nodo es una lista de candidatos descendientes del mismo nodo padre
# Soluciones es una lista de candidatos que podrían ser la solución
#------------------------------------------------

#------------------------------------------------
# Módulos
#------------------------------------------------

# Para manejar print como print() en Python 2.7
from __future__ import print_function

# Para leer la línea de órdenes
# Para consultar la versión de Python
import sys

# Para medir el tiempo que tarda el programa
import time

# Para generar combinaciones
import itertools

#------------------------------------------------
# Constantes
#------------------------------------------------

# Las cuatro operaciones
SUMA, RESTA, PRODUCTO, COCIENTE = range(4)

# Un número entero bastante grande
MAXVALOR = 1073741824

#------------------------------------------------
# Variables globales
#------------------------------------------------

# Los números datos
Datos = []

# El número meta
Meta = 0

# El mejor resultado obtenido
Mejor = 0

# La diferencia entre Meta y las posibles soluciones
Diferencia = MAXVALOR

# El menor número de operaciones para calcular la solución
MenorEsfuerzo = MAXVALOR

# La lista de soluciones que vamos encontrando
ListaSoluciones = []

# El número total de operaciones realizadas
TotalOperaciones = 0

#------------------------------------------------
# Funciones
#------------------------------------------------

#------------------------------------------------
# Resuelve el problema usando Datos y Meta
def Resuelve():
    global ListaSoluciones

    # Preparamos el nodo raíz
    NodoRaiz = []
    for Num in Datos:
        Candidato = (Num,str(Num),0)
        NodoRaiz.append(Candidato)

    # Lanzamos recursivamente la búsqueda de soluciones
    Busca(NodoRaiz)

    # Eliminamos soluciones obviamente repetidas
    ListaSoluciones = list(set(ListaSoluciones))

    # Nota: solo se usa la primera solución de la lista,
    #       pero se podrían estudiar todas para escoger
    #       la que parezca más adecuada

#------------------------------------------------
# Busca soluciones a partir de un nodo
def Busca(Nodo):
    global TotalOperaciones

    # Comprobamos cada candidato por si nos vale
    for Candidato in Nodo:
        Comprueba(Candidato)

    # Probamos todas las parejas de candidatos
    for Pareja in itertools.combinations(Nodo, 2):
        # Tomamos los datos de la pareja
        Candidato1 = Pareja[0]
        Candidato2 = Pareja[1]
        Num1 = Candidato1[0]
        Num2 = Candidato2[0]
        Cam1 = Candidato1[1]
        Cam2 = Candidato2[1]
        Esf1 = Candidato1[2]
        Esf2 = Candidato2[2]

        # Probamos todas las operaciones posibles
        for Operacion in range(4):
            HayQueCalcular = False

            if Operacion == SUMA:
                NuevoNum = Num1+Num2
                NuevoCam = Cam1 + ',' + Cam2 + ',+'
                HayQueCalcular = True

            if Operacion == RESTA:
                if Num1 > Num2:
                    NuevoNum = Num1-Num2
                    NuevoCam = Cam1 + ',' + Cam2 + ',-'
                    HayQueCalcular = True
                if Num1 < Num2:
                    NuevoNum = Num2-Num1
                    NuevoCam = Cam2 + ',' + Cam1 + ',-'
                    HayQueCalcular = True

            if Operacion == PRODUCTO:
                if Num1 != 1 and Num2 != 1:
                    NuevoNum = Num1*Num2
                    NuevoCam = Cam1 + ',' + Cam2 + ',*'
                    HayQueCalcular = True

            if Operacion == COCIENTE:
                if Num1 % Num2 == 0 and Num2 != 1:
                    NuevoNum = Num1//Num2
                    NuevoCam = Cam1 + ',' + Cam2 + ',/'
                    HayQueCalcular = True
                if Num2 % Num1 == 0 and Num1 != 1:
                    NuevoNum = Num2//Num1
                    NuevoCam = Cam2 + ',' + Cam1 + ',/'
                    HayQueCalcular = True

            if HayQueCalcular:
                # Una operación más
                TotalOperaciones += 1

                # Preparamos el nuevo candidato
                Esfuerzo = Esf1 + Esf2 + 1
                NuevoCandidato = (NuevoNum,NuevoCam,Esfuerzo)

                # Preparamos un nuevo nodo quitando la pareja
                # y añadiendo el nuevo candidato
                NuevoNodo = Nodo[:]
                NuevoNodo.remove(Candidato1)
                NuevoNodo.remove(Candidato2)
                NuevoNodo.append(NuevoCandidato)

                # Buscamos a partir del nuevo nodo
                Busca(NuevoNodo)

#------------------------------------------------
# Comprueba si hay que guardar un candidato
def Comprueba(Candidato):
    global Mejor, Diferencia, MenorEsfuerzo, ListaSoluciones

    # Tomamos los datos del candidato
    Numero = Candidato[0]
    Camino = Candidato[1]
    Esfuerzo = Candidato[2]
    DiferenciaPropuesta = abs(Meta-Numero)

    # Miramos si se podría guardar
    if DiferenciaPropuesta <= Diferencia:
        # Si es mejor candidato que los que teníamos, borramos lo anterior
        if DiferenciaPropuesta < Diferencia or Esfuerzo < MenorEsfuerzo:
            ListaSoluciones = []
            Mejor = Numero
            Diferencia = DiferenciaPropuesta
            MenorEsfuerzo = Esfuerzo
        # Lo guardamos si nos convence el esfuerzo
        if Esfuerzo <= MenorEsfuerzo:
            ListaSoluciones.append(Candidato)

#------------------------------------------------
# Explica un número encontrado
def ExplicaNumero(Num):
    Exacto = u' ¡Exacto!' if Num == Meta else ''
    Informa (u'Mejor aproximación encontrada: ' + str(Num) + Exacto)

#------------------------------------------------
# Explica una solución en RPN paso a paso
def ExplicaSolucion(Cadena):
    # La pila RPN para las operaciones
    Pila = []

    # Descomponemos la cadena
    ListaItems = Cadena.split(',')

    # Informamos
    Informa (u'Solución:')

    # Recorremos la lista de items
    for Item in ListaItems:

        # El ítem puede ser una operación
        if Item in '+-*/':
            # Los operandos
            Num2 = Pila.pop(0)
            Num1 = Pila.pop(0)

            # Realizamos la operación
            if Item == '+': Resultado = Num1+Num2
            if Item == '-': Resultado = Num1-Num2
            if Item == '*': Resultado = Num1*Num2
            if Item == '/': Resultado = Num1//Num2

            # Guardamos el resultado
            Pila.insert(0,Resultado)

            # Imprimimos la operación en notación algebraica usual
            Texto = str(Num1) + ' ' + Item + ' ' + \
                    str(Num2) + ' = ' + str(Resultado)
            Informa ('   ' + Texto)

        # El ítem puede ser un número
        else: Pila.insert(0,int(Item))

#------------------------------------------------
# Muesta los datos en pantalla
def ExplicaDatos():
    Informa (u'Números de partida: ' + ' '.join(map(str,Datos)))

#------------------------------------------------
# Muesta el objetivo en pantalla
def ExplicaMeta():
    Informa (u'Número objetivo: ' + str(Meta))

#------------------------------------------------
# Convierte una cantidad de tiempo a formato legible
def HumanizaTiempo(Tiempo):
    if Tiempo<60:
        ExpresionTiempo = str(Tiempo) + ' s'
    else:
        Minutos = Tiempo // 60
        Segundos = Tiempo % 60
        if Minutos<60:
            ExpresionTiempo = str(Minutos) + ' min ' + str(Segundos) + ' s'
        else:
            Horas = Minutos // 60
            Minutos = Minutos % 60
            ExpresionTiempo = (str(Horas) + ' h ' + str(Minutos) + ' min ' +
                               str(Segundos) + ' s')
    return ExpresionTiempo

#------------------------------------------------
# Escribe textos en pantalla
def Informa(Texto): print (Texto)

#------------------------------------------------
# Dice la versión de Python
def VersionPython():
    return '.'.join(map(str,[sys.version_info[i] for i in range(3)]))

#------------------------------------------------
# El programa
#------------------------------------------------

# Presentación
Informa ('Analizador aritmético exhaustivo')
Informa ('(Usando Python ' + VersionPython() + ')')

# Examinamos la línea de órdenes
NumPar = len(sys.argv)

# No hay suficientes parámetros
if NumPar < 3:
    Informa ('Uso: ./anaritex dato1 ... datoN Meta')

# Leemos los parámetros
else:
    # Los datos
    Datos = list(map(int, sys.argv[1:NumPar-1]))
    ExplicaDatos()

    # La meta
    Meta = int(sys.argv[NumPar-1])
    ExplicaMeta()

    # Resolvemos
    Informa ('Analizando...')
    Tiempo = time.time()
    Resuelve()
    Tiempo = int(time.time() - Tiempo + 0.5)
    ExpresionTiempo = HumanizaTiempo(Tiempo)

    # Informamos
    ExplicaNumero(ListaSoluciones[0][0])
    ExplicaSolucion(ListaSoluciones[0][1])
    Informa ('Operaciones analizadas: ' + str(TotalOperaciones))
    Informa ('Tiempo empleado: ' + ExpresionTiempo)
